#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <packet.h>
#include <preproc.h>
#include <decode.h>
#include <plugin.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* Stuff we import */
proc_preproc_activate preproc_activate;
proc_preproc_dispatch preproc_dispatch;
proc_decode_proto decode_proto;
struct proto *tcpproto=NULL;

/* These are optionally implemented, so don't check em */
u_int8_t omask=(TCP_ECE|TCP_CWR);

/* Mask to determine fragments */
unsigned short ipfmask=__constant_htons(IP_MF|IP_OFFMASK);

#define NONE		0
#define ESTABLISHED	1
#define SYN_SENT	2
#define SYN_RECV	3
#define FIN_WAIT_1	4
#define FIN_WAIT_2	5
#define TIME_WAIT	6
#define CLOSE		7
#define CLOSE_WAIT	8
#define LAST_ACK	9
#define TEARDOWN	10

/* Strings for our states */
char *state[]={
	"NONE",
	"ESTABLISHED",
	"SYN_SENT",
	"SYN_RECV",
	"FIN_WAIT_1",
	"FIN_WAIT_2",
	"TIME_WAIT",
	"CLOSE",
	"CLOSE_WAIT",
	"LAST_ACK",
	"RST"
};

/* Must be a power of 2 */
#define TCPHASH 256

/* TCP hashing function */
static inline unsigned int tcp_hashfn(
	u_int32_t saddr, u_int32_t daddr, 
	u_int16_t sport, u_int16_t dport)
{
	unsigned int h = ((daddr ^ dport) ^ (saddr ^ sport));
	h ^= h>>16;
	h ^= h>>8;
	return h & (TCPHASH - 1);
}

struct tcpstream {
	struct tcpstream *next;
	struct tcpstream **pprev;

	/* Network order */
	u_int32_t	saddr;
	u_int32_t	daddr;
	u_int16_t	sport;
	u_int16_t	dport;
	
	/* Host order */
	u_int32_t	isn;
	u_int32_t	seq;
	u_int32_t	ack;

	u_int16_t	state;
};
struct tcpstream *streams[TCPHASH];
unsigned int num_streams=0;
unsigned int max_streams=TCPHASH*512;

struct tcpstream *find_stream(u_int32_t saddr,
	u_int32_t daddr,
	u_int16_t sport,
	u_int16_t dport)
{
	struct tcpstream *t;

	for(t=streams[0]; t; t=t->next) {
		if ( t->saddr==saddr &&
			t->daddr==daddr &&
			t->sport==sport &&
			t->dport==dport ) {
			/* Move to front */
			if ( t->next )
				t->next->pprev=t->pprev;
			*t->pprev=t->next;
			if ( (t->next=streams[0]) ) 
				t->next->pprev=&t->next;
			streams[0]=t;
			t->pprev=&streams[0];
			return t;
		}
	}

	return NULL;
}

void transition(struct tcpstream *t, int st)
{
	struct in_addr s, d;
	char sip[32], dip[32];

	t->state=st;
	
	s.s_addr=t->saddr;
	d.s_addr=t->daddr;
	strncpy(sip, inet_ntoa(s), 32);
	strncpy(dip, inet_ntoa(d), 32);

	printf("tcpstream: state(%u): %s:%u -> %s:%u %s\n",
		num_streams,
		sip, ntohs(t->sport),
		dip, ntohs(t->dport),
		state[t->state]);
}

void tcp_packet(struct tcpstream *t, struct pkt_tcphdr *tcph)
{
	if ( t->state==NONE ) {
		if ( tcph->flags.bits.syn && !tcph->flags.bits.ack ) {
			/* Active open */
			transition(t, SYN_SENT);
		}
	}else if ( t->state==ESTABLISHED || t->state==SYN_RECV ) {
		if ( tcph->flags.bits.fin ) {
			transition(t, FIN_WAIT_1);
		}
	}else if ( t->state==CLOSE_WAIT ) {
		if ( tcph->flags.bits.fin ) {
			transition(t, LAST_ACK);
		}
	}
}

void tcp_reply(struct tcpstream *t, struct pkt_tcphdr *tcph)
{
	if ( t->state==SYN_RECV || t->state==SYN_SENT ) {
		if ( tcph->flags.bits.ack ) {
			transition(t, ESTABLISHED);
		}
	}else if ( t->state==ESTABLISHED ) {
		if ( tcph->flags.bits.fin) {
			transition(t, CLOSE_WAIT);
		}
	}else if ( t->state==LAST_ACK ) {
		if ( tcph->flags.bits.ack ) {
			transition(t, CLOSE);
		}
	}else if ( t->state==FIN_WAIT_1 ) {
		if ( tcph->flags.bits.fin && tcph->flags.bits.ack ) {
			transition(t, TIME_WAIT);
		}else if ( tcph->flags.bits.fin ) {
			transition(t, LAST_ACK);
		}else if ( tcph->flags.bits.ack ) {
			transition(t, FIN_WAIT_2);
		}
	}
}

/* Preprocessor entry point */
int tcpstream_process(struct packet *pkt, unsigned int i, void *priv)
{
	struct tcpstream *t, *t2;
	struct pkt_iphdr *iph;
	struct pkt_tcphdr *tcph;
	unsigned int teardown=0;
	
	if ( i+1 >= pkt->llen ) return PREPROC_OK;
	if ( pkt->layer[i+1].proto != tcpproto ) return PREPROC_OK;

	iph=pkt->layer[i].h.ip;
	tcph=pkt->layer[i+1].h.tcp;

	/* Don't bother with fragments */
	if ( iph->frag_off & ipfmask )
		return PREPROC_OK;

	if ( !(t=find_stream(iph->saddr, iph->daddr,
		tcph->sport, tcph->dport)) ) {
		if ( !tcph->flags.bits.syn ) {
			return PREPROC_DROP;
			printf("tcpstream: conn not seen\n");
		}

		if ( num_streams > max_streams ) {
			return PREPROC_OK;
			printf("tcpstream: max_streams reached\n");
		}

		if ( !(t=malloc(sizeof(*t))) ) {
			printf("tcpstream: oom\n");
			return PREPROC_OK;
		}

		num_streams++;

		/* Fill in the details */
		t->saddr=iph->saddr;
		t->daddr=iph->daddr;
		t->sport=tcph->sport;
		t->dport=tcph->dport;
		t->state=NONE;
		t->isn=ntohl(tcph->seq);
		t->seq=t->isn+1; /* Syn counts as 1 byte */
		t->ack=t->isn;
		
		if ( (t->next=streams[0]) ) 
			t->next->pprev=&t->next;
		streams[0]=t;
		t->pprev=&streams[0];
	}

	/* This needs to be "authenticated" */
	if ( tcph->flags.bits.rst ) {
		teardown=1;
	}

	/* Find reverse tuple */
	if ( (t2=find_stream(iph->daddr, iph->saddr,
		tcph->dport, tcph->sport)) ) {

		if ( teardown ) {
			transition(t2, TEARDOWN);
		}else{
			if ( tcph->flags.bits.ack && t2->state==SYN_SENT) {
				/* passive open */
				transition(t, SYN_RECV);
			}
			tcp_reply(t2, tcph);
		}
	}
	
	if ( teardown ) {
		transition(t, TEARDOWN);
	}else{
		tcp_packet(t, tcph);
	}

	/* Check whether to close the connections for good */
	if ( t2 ) {
		if ( teardown || (t->state==TIME_WAIT &&
			(t2->state==CLOSE || t2->state==TIME_WAIT)) ) {
			if ( t2->next )
				t2->next->pprev=t2->pprev;
			*t2->pprev=t2->next;
			free(t2);
			t2=NULL;
			num_streams--;
		}
	}

	if ( teardown || t->state==TIME_WAIT || t->state==CLOSE ) {
		if ( t->next )
			t->next->pprev=t->pprev;
		*t->pprev=t->next;
		free(t);
		t=NULL;
		num_streams--;
	}

	return PREPROC_OK;
}

int tcpstream_init(char *args)
{
	if ( tcpproto ) {
		printf("tcpstream: Can't add tcpstream twice!\n");
		return 0;
	}

	if ( !(tcpproto=decode_proto("tcp")) ) {
		printf("tcpstream: tcp protocol not loaded!\n");
		return 0;
	}
	
	/* TCP stream reassembly registering for IP realm?
	 * Well, we are only actually interested in TCP in IP */
	if (preproc_activate("ip", tcpstream_process, NULL)) {
		printf("tcpstream: TCP stream reassembly ENABLED.\n");
		return 1;
	}else{
		return 0;
	}
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	proc_preproc_add preproc_add;
	int ok=0;
	
	plugin_check(in, out);
	
	PLUGIN_ID("preproc.tcpstream", "TCP stream reassembly");
	PLUGIN_VERSION(0, 1);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");
	
	if ( !(preproc_add=in->import("preproc.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(preproc_activate=in->import("preproc.activate")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(preproc_dispatch=in->import("preproc.dispatch")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(decode_proto=in->import("decode.proto")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	ok+=preproc_add("tcpstream", tcpstream_init);

	memset(&streams, 0, sizeof(streams));
	
	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL;
}

int _PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
